import { __assign, __rest, __extends } from 'tslib';
import { removeVNodeDom, createVNode, createComponentVNode, mount, findDomFromVNode, patch as patch$1, unmount as unmount$1, Component as Component$1, callAll } from 'intact';
export * from 'intact';
import { createApp, h, KeepAlive, getCurrentInstance, cloneVNode, Comment, Text, camelize, isVNode, Fragment, vShow, isRef, createVNode as createVNode$1 } from 'vue';

// we must use this hack method to get patch function
var internals;
createApp({
    render: function () {
        return h(KeepAlive, null, h(function () {
            var instance = getCurrentInstance();
            internals = instance.parent.ctx.renderer;
        }));
    }
}).mount(document.createElement('div'));
var _a$1 = internals, patch = _a$1.p, unmount = _a$1.um;
var Wrapper = /*#__PURE__*/ /*#__PURE__*/ /** @class */ (function () {
    function Wrapper($props, $vNode, $SVG, $mountedQueue, $senior) {
        this.$props = $props;
        this.$vNode = $vNode;
        this.$SVG = $SVG;
        this.$mountedQueue = $mountedQueue;
        this.$senior = $senior;
        this.$inited = true;
        this.$lastInput = createVNode('div');
    }
    Wrapper.prototype.$init = function (props) { };
    Wrapper.prototype.$render = function (lastVNode, vNode, parentDom, anchor) {
        if (lastVNode) {
            removeVNodeDom(lastVNode, parentDom);
        }
        else if (!parentDom) {
            parentDom = document.createDocumentFragment();
        }
        var vnode = getVueVNode(vNode);
        patch(null, vnode, parentDom, anchor, getParent(this), null, this.$SVG);
        // add dom to the $lastInput for findDomFromVNode
        this.$lastInput.dom = vnode.el;
    };
    Wrapper.prototype.$update = function (lastVNode, vNode, parentDom, anchor) {
        var lastVnode = lastVNode.props.vnode;
        var nextVnode = getVueVNode(vNode);
        patch(lastVnode, nextVnode, parentDom, anchor, getParent(this), null, this.$SVG);
        this.$lastInput.dom = nextVnode.el;
    };
    Wrapper.prototype.$unmount = function (vNode, nextVNode) {
        unmount(vNode.props.vnode, getParent(this), null, !!nextVNode);
    };
    return Wrapper;
}());
function getParent(instance) {
    var $senior = instance.$senior;
    do {
        var vueInstance = $senior.vueInstance;
        if (vueInstance) {
            return vueInstance;
        }
    } while ($senior = $senior.$senior);
    // should not hit this
    /* istanbul ignore next */
    return null;
}
function getVueVNode(vNode) {
    var props = vNode.props;
    var vnode = props.vnode;
    // if we are reusing the vNode, clone it
    var hasCloned = false;
    if (vnode.el) {
        vnode = cloneVNode(vnode);
        hasCloned = true;
    }
    var shouldAssign = true;
    for (var key in props) {
        if (key === 'vnode')
            continue;
        // clone the vnode at first
        if (!hasCloned) {
            vnode = cloneVNode(vnode);
            hasCloned = true;
        }
        var _props = vnode.props;
        if (shouldAssign) {
            // props may be a EMPTY_OBJ, but we can not get its reference,
            // so we use a flag to handle it
            _props = vnode.props = __assign({}, _props);
            shouldAssign = false;
            // should change patchFlag to let Vue full diff props
            vnode.patchFlag |= 16;
        }
        var value = props[key];
        // is event
        if (key === 'className') {
            _props.class = value;
            // } else if (key === 'style') {
            // _props.style = {..._props.style, ...value};
        }
        else if (key.substr(0, 3) === 'ev-') {
            var name_1 = key.substr(3);
            _props["on" + name_1[0].toUpperCase() + name_1.substr(1)] = value;
        }
        else {
            _props[key] = value;
        }
    }
    vNode.props = __assign(__assign({}, props), { vnode: vnode });
    return vnode;
}

var toString = Object.prototype.toString;
var isArray = Array.isArray || function (arr) {
    return toString.call(arr) === '[object Array]';
};
function isInvalid(o) {
    return o === null || o === false || o === true || o === undefined;
}
function isStringOrNumber(o) {
    var type = typeof o;
    return type === 'string' || type === 'number';
}
function isFunction(o) {
    return typeof o === 'function';
}
function isString(o) {
    return typeof o === 'string';
}
var hasOwn = Object.prototype.hasOwnProperty;
var EMPTY_OBJ = {};
if (process.env.NODE_ENV !== 'production') {
    Object.freeze(EMPTY_OBJ);
}

function normalize(vnode) {
    if (isInvalid(vnode))
        return null;
    if (isStringOrNumber(vnode) || isIntactVNode(vnode))
        return vnode;
    var type = vnode.type;
    var vNode;
    if (isIntactComponent(vnode)) {
        var props = normalizeProps(vnode);
        vNode = createComponentVNode(4, type.Component, props, vnode.key, props.ref);
    }
    else {
        // ignore comment vNode
        if (type === Comment)
            return null;
        if (type === Text)
            return vnode.children;
        // spread fragment
        // if (type === Fragment) {
        // return normalizeChildren(vnode.children as VNodeArrayChildren);
        // }
        vNode = createComponentVNode(4, Wrapper, { vnode: vnode }, vnode.key);
    }
    // tell Vue that this is a read only object, and don't make it reactive
    vNode.__v_isReadonly = true;
    return vNode;
}
function isIntactComponent(vnode) {
    return !!vnode.type.Component;
}
function isIntactVNode(vNode) {
    return 'childrenType' in vNode;
}
function normalizeChildren(vNodes) {
    var loop = function (vNodes) {
        if (Array.isArray(vNodes)) {
            var ret_1 = [];
            vNodes.forEach(function (vNode) {
                if (Array.isArray(vNode)) {
                    ret_1.push.apply(ret_1, loop(vNode));
                }
                else if (isVNode(vNode)) {
                    ret_1.push(normalize(vNode));
                }
            });
            return ret_1;
        }
        return normalize(vNodes);
    };
    var ret = loop(vNodes);
    if (Array.isArray(ret)) {
        var l = ret.length;
        return l === 0 ? undefined : l === 1 ? ret[0] : ret;
    }
    return ret;
}
function normalizeProps(vnode) {
    var attrs = vnode.props;
    var slots = vnode.children;
    var Component = vnode.type.Component;
    var props = {};
    var propTypes = Component.typeDefs;
    for (var key in attrs) {
        var value = attrs[key];
        switch (key) {
            case 'ref': break;
            case 'class':
                props.className = value;
                break;
            case 'modelValue':
                props.value = value;
                break;
            default:
                if (isOn(key)) {
                    normalizeEvents(props, key, value);
                    break;
                }
                if (propTypes) {
                    var camelizedKey = camelize(key);
                    if (hasOwn.call(propTypes, camelizedKey)) {
                        value = normalizeBoolean(propTypes, key, camelizedKey, value);
                        key = camelizedKey;
                    }
                }
                props[key] = value;
                break;
        }
    }
    normalizeSlots(slots, props);
    normalizeDirs(vnode.dirs, props);
    normalizeRef(vnode.ref, props);
    return props;
}
var onRE = /^on[^a-z]/;
var isOn = function (key) { return onRE.test(key); };
function normalizeEvents(props, key, value) {
    var name;
    var cb = value;
    var _isArray = isArray(value);
    var changeCallback = function (propName) { return function (v) {
        var modifiersKey = (propName === 'value' ? 'model' : propName) + "Modifiers";
        var _a = props[modifiersKey] || EMPTY_OBJ, number = _a.number, trim = _a.trim;
        if (trim) {
            v = String(v).trim();
        }
        else if (number) {
            v = Number(v);
        }
        if (_isArray) {
            value.forEach(function (value) { return value(v); });
        }
        else {
            value(v);
        }
    }; };
    if (key.startsWith('onUpdate:')) {
        var propName = camelize(key.substr(9));
        if (propName === 'modelValue')
            propName = 'value';
        name = "$model:" + propName;
        cb = changeCallback(propName);
    }
    else {
        key = key.substring(2);
        if (key.startsWith('Change:')) {
            // onChange:name -> ev-$change:name
            name = "$change:" + key.substring(7);
        }
        else {
            // onClick -> ev-click
            // onChange -> ev-change
            name = key[0].toLowerCase() + key.substring(1);
        }
        if (_isArray) {
            cb = function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                value.forEach(function (value) { return value.apply(void 0, args); });
            };
        }
    }
    props["ev-" + name] = cb;
}
function normalizeBoolean(propTypes, key, camelizedKey, value) {
    var tmp;
    if ((
    // value is Boolean
    (tmp = propTypes[camelizedKey]) === Boolean ||
        tmp && (
        // value contains Boolean
        isArray(tmp) && tmp.indexOf(Boolean) > -1 ||
            (tmp = tmp.type) && (
            // value.type is Boolean
            tmp === Boolean ||
                // value.type contains Boolean
                isArray(tmp) && tmp.indexOf(Boolean) > -1))) &&
        (value === '' || value === key)) {
        value = true;
    }
    return value;
}
function normalizeSlots(slots, props) {
    if (!slots)
        return;
    // is array children
    if (isArray(slots)) {
        return props.children = normalizeChildren(slots);
    }
    // is string
    if (isStringOrNumber(slots)) {
        return props.children = slots;
    }
    // the default slot maybe a scope slot, but we can not detect
    // whether it is or not, so we try to normalize it as children and
    // then treat it as a default scope slot too.
    if (slots.default) {
        var slot = slots.default;
        try {
            // Vue will detect whether the slot is invoked outside or not,
            // but it does not affetch anything in here,
            // so we keep the warning silent
            //
            // Vue will warn if we get property of undefined, we keep it silent
            // silentWarn();
            var validSlotContent = ensureValidVNode(slot());
            props.children = validSlotContent ? normalizeChildren(validSlotContent) : null;
            // resetWarn();
        }
        catch (e) { }
    }
    var blocks = null;
    var _loop_1 = function (key) {
        if (isInternalKey(key))
            return "continue";
        var slot = slots[key];
        if (!blocks)
            blocks = {};
        blocks[key] = function (parent, data) {
            // if the content is invalid, use parent instead
            // this is the default behavior of Vue
            var validSlotContent = ensureValidVNode(slot(data));
            return validSlotContent ? normalizeChildren(validSlotContent) : parent();
        };
    };
    for (var key in slots) {
        _loop_1(key);
    }
    if (blocks) {
        props.$blocks = blocks;
    }
}
function ensureValidVNode(vnodes) {
    if (!isArray(vnodes))
        vnodes = [vnodes];
    return vnodes.some(function (child) {
        if (!isVNode(child)) {
            return true;
        }
        if (child.type === Comment) {
            return false;
        }
        if (child.type === Fragment && !ensureValidVNode(child.children)) {
            return false;
        }
        return true;
    }) ? vnodes : null;
}
var isInternalKey = function (key) { return key[0] === '_' || key === '$stable'; };
function normalizeDirs(dirs, props) {
    if (!dirs)
        return;
    dirs.find(function (_a) {
        var dir = _a.dir, value = _a.value;
        // only handle v-show
        if (dir === vShow) {
            if (!value) {
                (props.style || (props.style = {})).display = 'none';
            }
            return true;
        }
    });
}
function normalizeRef(rawRef, props) {
    if (isFunction(rawRef))
        props.ref = rawRef;
    else if (rawRef) {
        props.ref = function (i) {
            setRef(rawRef, i);
        };
    }
}
function setRef(rawRef, value) {
    if (isArray(rawRef)) {
        rawRef.forEach(function (r, i) { return setRef(r, value); });
        return;
    }
    var owner = rawRef.i, ref = rawRef.r;
    var refs = !Object.keys(owner.refs).length ? (owner.refs = {}) : owner.refs;
    if (isString(ref)) {
        refs[ref] = value;
    }
    else if (isRef(ref)) {
        ref.value = value;
    }
    else if (isFunction(ref)) {
        ref(value, refs);
    }
    else {
        console.warn('Invalid template ref type:', value, "(" + typeof value + ")");
    }
}

function functionalWrapper(Component) {
    function Ctor(props, context) {
        if (context) {
            // invoked by Vue
            var forwardRef = props.forwardRef, rest = __rest(props, ["forwardRef"]);
            var _props = normalizeProps({
                props: rest,
                children: context.slots,
                type: {
                    Component: Component,
                },
                ref: forwardRef,
            });
            // functional component of intact must return VNodeComponentClass
            var vNode = Component(_props, true /* is in vue */);
            if (Array.isArray(vNode)) {
                return vNode.map(function (vNode) { return toVueVNode(vNode); });
            }
            return toVueVNode(vNode);
        }
        else {
            // invoked by Intact
            return Component(props);
        }
    }
    Ctor.displayName = Component.displayName || Component.name;
    return Ctor;
}
function toVueVNode(vNode) {
    // if (isStringOrNumber(vNode)) return vNode;
    if (vNode) {
        var props = vNode.props;
        return h(vNode.tag, 
        // because Vue will normalize some styles, but the props of vNode in Intact
        // is immutable, we must clone it here.
        props ? __assign({}, props) : null);
    }
}

function setScopeId(el, vnode, scopeId, slotScopeIds, parentComponent) {
    if (el.nodeType !== 1)
        return;
    if (scopeId) {
        hostSetScopeId(el, scopeId);
    }
    /* istanbul ignore next */
    if (slotScopeIds) {
        for (var i = 0; i < slotScopeIds.length; i++) {
            hostSetScopeId(el, slotScopeIds[i]);
        }
    }
    if (parentComponent) {
        var subTree = parentComponent.subTree;
        if (vnode === subTree) {
            var parentVNode = parentComponent.vnode;
            setScopeId(el, parentVNode, parentVNode.scopeId, 
            // slotScopeIds is @internal
            parentVNode.slotScopeIds, parentComponent.parent);
        }
    }
}
function hostSetScopeId(el, scopeId) {
    el.setAttribute(scopeId, '');
}

var currentInstance = null;
var _a = createStack(), pushMountedQueue = _a[0], popMountedQueue = _a[1];
var Component = /*#__PURE__*/ /*#__PURE__*/ /** @class */ (function (_super) {
    __extends(Component, _super);
    function Component(props, $vNode, $SVG, $mountedQueue, $parent) {
        var _this = _super.call(this, props, $vNode, $SVG, $mountedQueue, $parent) || this;
        _this.$isVueNext = false;
        var vuePublicInstance = $vNode._vueInstance;
        _this.vueInstance = vuePublicInstance;
        if (vuePublicInstance) {
            // set the instance to the setupState of vueIntance.$
            // ps: setupState is @internal in vue
            vuePublicInstance.setupState.instance = _this;
            _this.$isVueNext = true;
        }
        // disable async component 
        _this.$inited = true;
        return _this;
    }
    Object.defineProperty(Component, "__vccOpts", {
        get: function () {
            var Ctor = this;
            if (Ctor.__cache) {
                return Ctor.__cache;
            }
            return (Ctor.__cache = {
                name: Ctor.displayName || Ctor.name,
                Component: Ctor,
                // Don't inherit attributes, becasue kpc component has do it.
                // https://v3.vuejs.org/guide/component-attrs.html#attribute-inheritance
                inheritAttrs: false,
                setup: function (props, setupContext) {
                    var setupState = { instance: null };
                    var proxy = new Proxy(setupState, {
                        get: function (_a, key) {
                            var instance = _a.instance;
                            if (key === '__v_isReactive')
                                return true;
                            if (instance === null)
                                return null;
                            if (key === 'instance')
                                return instance;
                            var value = instance[key];
                            if (isFunction(value)) {
                                // should bind instance, otherwise the `this` may point to proxyToUse
                                return value.bind(instance);
                            }
                            return value;
                        },
                        set: function (setupState, key, value) {
                            // if (key === 'instance') {
                            return Reflect.set(setupState, key, value);
                            // }
                            // return Reflect.set(setupState.instance!, key, value);
                        },
                        getOwnPropertyDescriptor: function () {
                            return {
                                value: undefined,
                                writable: true,
                                enumerable: true,
                                configurable: true,
                            };
                        },
                        ownKeys: function () {
                            return [];
                        }
                    });
                    return proxy;
                },
                render: function (proxyToUse, renderCache, props, setupState) {
                    var vueInstance = proxyToUse.$;
                    var vNode = normalize(vueInstance.vnode);
                    var _setScopeId = function (element) {
                        var vnode = vueInstance.vnode;
                        setScopeId(element, vnode, vnode.scopeId, vnode.slotScopeIds, vueInstance.parent);
                    };
                    var mountedQueue = pushMountedQueue([]);
                    var parentComponent = getIntactParent(vueInstance.parent);
                    var isSVG = parentComponent ? parentComponent.$SVG : false;
                    var vnode = createVNode$1(Comment, { key: '1' });
                    var subTree = Ctor.$doubleVNodes ?
                        createVNode$1(Fragment, null, [vnode, createVNode$1(Comment, { key: '2' })]) :
                        vnode;
                    if (!vueInstance.isMounted) {
                        // add subTree firstly, because when we mount vue element in intact,
                        // Vue need the property on calling setScopeId
                        vueInstance.subTree = subTree;
                        vNode._vueInstance = vueInstance;
                        mount(vNode, null, parentComponent, isSVG, null, mountedQueue);
                        // hack the nodeOps of Vue to create the real dom instead of a comment
                        var elements_1 = [findDomFromVNode(vNode, true)];
                        var index_1 = 0;
                        if (Ctor.$doubleVNodes) {
                            elements_1.push(findDomFromVNode(vNode, false));
                        }
                        var nativeCreateComment_1 = document.createComment;
                        document.createComment = function () {
                            var element = elements_1[index_1];
                            if (++index_1 === elements_1.length) {
                                document.createComment = nativeCreateComment_1;
                            }
                            // scope id
                            _setScopeId(element);
                            return element;
                        };
                    }
                    else {
                        var instance = setupState.instance;
                        var lastVNode = instance.$vNode;
                        patch$1(lastVNode, vNode, this.$el.parentElement, parentComponent, isSVG, null, mountedQueue, false);
                        // element may have chagned
                        // only check if the component returns only one vNode
                        if (!Ctor.$doubleVNodes) {
                            var element = findDomFromVNode(vNode, true);
                            var oldSubTree = vueInstance.subTree;
                            if (oldSubTree.el !== element) {
                                oldSubTree.el = element;
                                // set scope id
                                _setScopeId(element);
                            }
                        }
                    }
                    return subTree;
                },
                mounted: function () {
                    callMountedQueue();
                },
                updated: function () {
                    callMountedQueue();
                },
                beforeUnmount: function () {
                    // we should get property by instance, if the name starts with '$'
                    unmount$1(this.instance.$vNode, null);
                },
            });
        },
        enumerable: false,
        configurable: true
    });
    Component.prototype.$render = function (lastVNode, nextVNode, parentDom, anchor, mountedQueue) {
        var popInstance = pushInstance(this);
        _super.prototype.$render.call(this, lastVNode, nextVNode, parentDom, anchor, mountedQueue);
        popInstance();
    };
    Component.prototype.$update = function (lastVNode, nextVNode, parentDom, anchor, mountedQueue, force) {
        var popInstance = pushInstance(this);
        _super.prototype.$update.call(this, lastVNode, nextVNode, parentDom, anchor, mountedQueue, force);
        popInstance();
    };
    Component.$cid = 'IntactVueNext';
    Component.$doubleVNodes = false;
    Component.__cache = null;
    Component.functionalWrapper = functionalWrapper;
    Component.normalize = normalizeChildren;
    return Component;
}(Component$1));
function createStack() {
    var stack = [];
    function pushStack(item) {
        stack.push(item);
        return item;
    }
    function popStack() {
        return stack.pop();
    }
    return [pushStack, popStack];
}
function callMountedQueue() {
    var mountedQueue = popMountedQueue();
    /* istanbul ignore next */
    if (process.env.NODE_ENV !== 'production') {
        if (!mountedQueue) {
            throw new Error("\"mountedQueue\" is undefined, maybe this is a bug of Intact-Vue");
        }
    }
    callAll(mountedQueue);
}
var _b = createStack(), _pushInstance = _b[0], _popInstance = _b[1];
function pushInstance(instance) {
    var lastIntance = currentInstance;
    currentInstance = _pushInstance(instance);
    return function () {
        _popInstance();
        currentInstance = lastIntance;
    };
}
function getIntactParent(parent) {
    if (currentInstance) {
        return currentInstance;
    }
    // maybe we mount/update a intact component in Vue component
    // let parent: ComponentInternalInstance | null = instance;
    while (parent) {
        var instance = parent.setupState.instance;
        if (instance instanceof Component) {
            return instance;
        }
        parent = parent.parent;
    }
    return null;
}

export { Component };
